from tensorflow.keras import Model
import numpy as np

from hw2_utils import Splits, TestCase, timeit, load_mnist, accuracy, run_tests


class TestMNIST(TestCase):
    def __init__(self, *args, **kwargs):
        self.module_name = "hw2_mnist"
        TestCase.__init__(self, *args, **kwargs)

    def setUp(self):
        TestCase.setUp(self)

        self.splits = Splits(
            *[split._replace(
                X=np.reshape(split.X, (split.X.shape[0], 28, 28, 1))
            ) for split in load_mnist()]
        )

    def test01_build_model_mnist1(self):
        """Make sure build_model_mnist1 returns the right thing."""

        model = self.solution.build_model_mnist1()

        self.assertIsInstance(model, Model)

        print(model.summary())

    def test02_train_model_mnist1(self):
        """Evaluate train_model_mnist1."""

        model = self.solution.build_model_mnist1()

        model, train_time = timeit(
            lambda: self.solution.train_model_mnist1(
                model,
                self.splits.train.X,
                self.splits.train.Y)
        )

        self.__class__.time_mnist1_train = train_time

        score_train = accuracy(
            model,
            self.splits.train.X, self.splits.train.Y
        )
        score_test = accuracy(
            model,
            self.splits.test.X, self.splits.test.Y
        )

        self.__class__.score_mnist1_train = score_train
        self.__class__.score_mnist1_test = score_test

        self.assertGreater(
            score_train, 0.9,
            msg=f"Model's train accuracy insufficient ({score_train} < 0.9)."
        )
        self.assertGreater(
            score_test, 0.9,
            msg=f"Model's test accuracy insufficient ({score_test} < 0.9)."
        )

    def test03_build_model_mnist2(self):
        """Make sure build_model_mnist2 returns the right thing."""

        model = self.solution.build_model_mnist2()

        self.assertIsInstance(model, Model)

        print(model.summary())

    def test04_train_model_mnist2(self):
        """Evaluate train_model_mnist2."""

        model = self.solution.build_model_mnist2()

        model, train_time = timeit(
            lambda: self.solution.train_model_mnist2(
                model,
                self.splits.train.X,
                self.splits.train.Y)
        )

        self.__class__.time_mnist2_train = train_time

        score_train = accuracy(
            model, self.splits.train.X, self.splits.train.Y
        )
        score_test = accuracy(
            model, self.splits.test.X, self.splits.test.Y
        )

        self.__class__.score_mnist2_train = score_train
        self.__class__.score_mnist2_test = score_test

        self.assertGreater(
            score_train, 0.9,
            msg=f"Model's train accuracy insufficient ({score_train} < 0.9)."
        )
        self.assertGreater(
            score_test, 0.9,
            msg=f"Model's test accuracy insufficient ({score_test} < 0.9)."
        )

    def test05_improvement(self):
        """Check whether second model improved the first."""

        self.assertTrue(hasattr(self.__class__, "score_mnist1_test"),
                        msg="First model failed to train.")

        self.assertTrue(hasattr(self.__class__, "score_mnist2_test"),
                        msg="Second model failed to train.")

        print("First model has " \
              + f"accuracy {self.__class__.score_mnist1_test} " \
              + f"over time {self.__class__.time_mnist1_train}")

        print("Second model has " \
              + f" Accuracy {self.__class__.score_mnist2_test} " \
              + f"over time {self.__class__.time_mnist2_train}")

        self.assertGreater(
            self.__class__.score_mnist1_test, 0.9,
            msg="First model's test accuracy insufficient for comparison."
        )

        no_worse_accuracy = \
            self.__class__.score_mnist2_test \
            >= self.__class__.score_mnist1_test * 0.999

        better_accuracy = \
            self.__class__.score_mnist2_test \
            >= self.__class__.score_mnist1_test * 1.05

        no_worse_time = \
            self.__class__.time_mnist2_train * 0.999 \
            <= self.__class__.time_mnist1_train

        better_time = \
            self.__class__.time_mnist2_train * 1.05 \
            <= self.__class__.time_mnist1_train

        self.assertTrue(
            (no_worse_accuracy and better_time) or
            (no_worse_time and better_accuracy),
            msg="Second model did not offer required improvement." \
            + f" Accuracy {self.__class__.score_mnist2_test} " \
            + f"over time {self.__class__.time_mnist2_train} from " \
            + f"accuracy {self.__class__.score_mnist1_test} " \
            + f"over time {self.__class__.time_mnist1_train}"
        )


if __name__ == '__main__':
    run_tests(TestMNIST)
